# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utility functions and for loading morphs."""

import logging
import os
import shutil
import tempfile
import uuid
from functools import lru_cache
from io import StringIO, open
from pathlib import Path

import morphio
from neurom.core.morphology import Morphology
from neurom.core.population import Population
from neurom.exceptions import NeuroMError
from neurom.utils import warn_deprecated

L = logging.getLogger(__name__)


def _is_morphology_file(filepath):
    """Check if `filepath` is a file with one of morphology file extensions."""
    return filepath.is_file() and filepath.suffix.lower() in {'.swc', '.h5', '.asc'}


class MorphLoader:
    """Caching morphology loader.

    Arguments:
        directory: path to directory with morphology files
        file_ext: file extension to look for (if not set, will pick any of .swc|.h5|.asc)
        cache_size: size of LRU cache (if not set, no caching done)
    """

    def __init__(self, directory, file_ext=None, cache_size=None):
        """Initialize a MorphLoader object."""
        self.directory = Path(directory)
        self.file_ext = file_ext
        if cache_size is not None:
            self.get = lru_cache(maxsize=cache_size)(self.get)

    def _filepath(self, name):
        """File path to `name` morphology file."""
        if self.file_ext is None:
            candidates = self.directory.glob(name + ".*")
            try:
                return next(filter(_is_morphology_file, candidates))
            except StopIteration as e:
                raise NeuroMError("Can not find morphology file for '%s' " % name) from e
        else:
            return Path(self.directory, name + self.file_ext)

    # pylint:disable=method-hidden
    def get(self, name):
        """Get `name` morphology data."""
        return load_morphology(self._filepath(name))


def get_morph_files(directory):
    """Get a list of all morphology files in a directory.

    Returns:
        list with all files with extensions '.swc' , 'h5' or '.asc' (case insensitive)
    """
    directory = Path(directory)
    return list(filter(_is_morphology_file, directory.iterdir()))


def get_files_by_path(path):
    """Get a file or set of files from a file path.

    Return list of files with path
    """
    path = Path(path)
    if path.is_file():
        return [path]
    if path.is_dir():
        return get_morph_files(path)

    raise IOError('Invalid data path %s' % path)


def _get_file(stream, extension):
    """Returns the filename of the file to read."""
    if isinstance(stream, str):
        stream = StringIO(stream)
    fd, temp_file = tempfile.mkstemp(suffix=f'{uuid.uuid4()}.{extension}', prefix='neurom-')
    os.close(fd)
    with open(temp_file, 'w') as fd:
        stream.seek(0)
        shutil.copyfileobj(stream, fd)
    return temp_file


def load_morphology(morph, reader=None):
    """Build section trees from a morphology or a h5, swc or asc file.

    Args:
        morph (str|Path|Morphology|morphio.Morphology|morphio.mut.Morphology): a morphology
            representation. It can be:

            - a filename with the h5, swc or asc extension
            - a NeuroM Neuron object
            - a morphio mutable or immutable Morphology object
            - a stream that can be put into a io.StreamIO object. In this case, the READER argument
              must be passed with the corresponding file format (asc, swc and h5)
        reader (str): Optional, must be provided if morphology is a stream to
                      specify the file format (asc, swc, h5)

    Returns:
        A Morphology object

    Examples::

            morphology = neurom.load_morphology('my_morphology_file.h5')
            morphology = neurom.load_morphology(morphio.Morphology('my_morphology_file.h5'))
            morphology = nm.load_morphology(io.StringIO('''((Dendrite)
                                                   (3 -4 0 2)
                                                   (3 -6 0 2)
                                                   (3 -8 0 2)
                                                   (3 -10 0 2)
                                                   (
                                                     (0 -10 0 2)
                                                     (-3 -10 0 2)
                                                     |
                                                     (6 -10 0 2)
                                                     (9 -10 0 2)
                                                   )
                                                   )'''), reader='asc')
    """
    if isinstance(morph, (Morphology, morphio.Morphology, morphio.mut.Morphology)):
        return Morphology(morph)

    if reader:
        return Morphology(_get_file(morph, reader))

    return Morphology(morph, Path(morph).name)


def load_neuron(morph, reader=None):
    """Deprecated in favor of ``load_morphology``."""
    warn_deprecated('`neurom.io.utils.load_neuron` is deprecated in favor of '
                    '`neurom.io.utils.load_morphology`')  # pragma: no cover
    return load_morphology(morph, reader)  # pragma: no cover


def load_morphologies(morphs,
                      name=None,
                      ignored_exceptions=(),
                      cache=False):
    """Create a population object.

    From all morphologies in a directory of from morphologies in a list of file names.

    Arguments:
        morphs(str|Path|Iterable[Path]): path to a folder or list of paths to morphology files
        name (str): optional name of population. By default 'Population' or\
            filepath basename depending on whether morphologies is list or\
            directory path respectively.
        ignored_exceptions (tuple): NeuroM and MorphIO exceptions that you want to ignore when
            loading morphologies
        cache (bool): whether to cache the loaded morphologies in memory

    Returns:
        Population: population object
    """
    if isinstance(morphs, (str, Path)):
        files = get_files_by_path(morphs)
        name = name or Path(morphs).name
    else:
        files = morphs
        name = name or 'Population'
    return Population(files, name, ignored_exceptions, cache)


def load_neurons(morphs, name=None, ignored_exceptions=(), cache=False):
    """Deprecated in favor of ``load_morphologies``."""
    warn_deprecated('`neurom.io.utils.load_neurons` is deprecated in favor of '
                    '`neurom.io.utils.load_morphologies`')  # pragma: no cover
    return load_morphologies(morphs, name, ignored_exceptions, cache)  # pragma: no cover



"""For backward compatibility only."""
# pylint: skip-file

from neurom.check.morphology_checks import *  # pragma: no cover
from neurom.utils import deprecated_module  # pragma: no cover

deprecated_module('Module `neurom.check.neuron_checks` is deprecated. Use'
                  '`neurom.check.morphology_checks` instead.')  # pragma: no cover



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Statistics for morphologies."""

import csv
import json
import logging
import multiprocessing
import os
import warnings
from collections import defaultdict
from collections.abc import Sized
from functools import partial
from itertools import chain, product
from pathlib import Path
import pkg_resources
from tqdm import tqdm
import numpy as np
import pandas as pd
from morphio import SomaError

import neurom as nm
from neurom.apps import get_config
from neurom.core.morphology import Morphology
from neurom.exceptions import ConfigError
from neurom.features import _NEURITE_FEATURES, _MORPHOLOGY_FEATURES, _POPULATION_FEATURES, \
    _get_feature_value_and_func
from neurom.io.utils import get_files_by_path
from neurom.utils import NeuromJSON, warn_deprecated

L = logging.getLogger(__name__)

EXAMPLE_CONFIG = Path(pkg_resources.resource_filename('neurom.apps', 'config'), 'morph_stats.yaml')
IGNORABLE_EXCEPTIONS = {'SomaError': SomaError}


def _run_extract_stats(morph, config):
    """The function to be called by multiprocessing.Pool.imap_unordered."""
    if not isinstance(morph, Morphology):
        morph = nm.load_morphology(morph)
    return morph.name, extract_stats(morph, config)


def extract_dataframe(morphs, config, n_workers=1):
    """Extract stats grouped by neurite type from morphs.

    Arguments:
        morphs: a morphology, population, neurite tree or list of morphology paths
        config (dict): configuration dict. The keys are:
            - neurite_type: a list of neurite types for which features are extracted
              If not provided, all neurite_type will be used
            - neurite: a dictionary {{neurite_feature: mode}} where:
                - neurite_feature is a string from NEURITEFEATURES or NEURONFEATURES
                - mode is an aggregation operation provided as a string such as:
                  ['min', 'max', 'median', 'mean', 'std', 'raw', 'sum']
            - morphology: same as neurite entry, but it will not be run on each neurite_type,
              but only once on the whole morphology.
        n_workers (int): number of workers for multiprocessing (on collection of morphs)

    Returns:
        The extracted statistics

    Note:
        An example config can be found at:

    {config_path}
    """
    if isinstance(morphs, Morphology):
        morphs = [morphs]
    config = config.copy()

    func = partial(_run_extract_stats, config=config)
    if n_workers == 1:
        stats = list(map(func, morphs))
    else:
        if n_workers > os.cpu_count():
            warnings.warn(f'n_workers ({n_workers}) > os.cpu_count() ({os.cpu_count()}))')
        with multiprocessing.Pool(n_workers) as pool:
            stats = list(pool.imap(func, morphs))

    columns = [('property', 'name')] + [
        (key1, key2) for key1, data in stats[0][1].items() for key2 in data
    ]
    rows = [[name] + list(chain.from_iterable(features.values() for features in data.values()))
            for name, data in stats]
    return pd.DataFrame(columns=pd.MultiIndex.from_tuples(columns), data=rows)


extract_dataframe.__doc__ = extract_dataframe.__doc__.format(config_path=EXAMPLE_CONFIG)


def _get_feature_stats(feature_name, morphs, modes, kwargs):
    """Insert the stat data in the dict.

    If the feature is 2-dimensional, the feature is flattened on its last axis
    """
    data = {}
    value, func = _get_feature_value_and_func(feature_name, morphs, **kwargs)
    shape = func.shape
    if len(shape) > 2:
        raise ValueError(f'Len of "{feature_name}" feature shape must be <= 2')  # pragma: no cover

    for mode in modes:
        stat_name = f'{mode}_{feature_name}'

        stat = value
        if isinstance(value, Sized):
            if len(value) == 0 and mode not in {'raw', 'sum'}:
                stat = None
            else:
                stat = getattr(np, mode)(value, axis=0)

        if len(shape) == 2:
            for i in range(shape[1]):
                data[f'{stat_name}_{i}'] = stat[i] if stat is not None else None
        else:
            data[stat_name] = stat
    return data


def extract_stats(morphs, config):
    """Extract stats from morphs.

    Arguments:
        morphs: a morphology, population, neurite tree or list of morphology paths/str
        config (dict): configuration dict. The keys are:
            - neurite_type: a list of neurite types for which features are extracted
              If not provided, all neurite_type will be used.
            - neurite: a dictionary {{neurite_feature: mode}} where:
                - neurite_feature is a string from NEURITEFEATURES or NEURONFEATURES
                - mode is an aggregation operation provided as a string such as:
                  ['min', 'max', 'median', 'mean', 'std', 'raw', 'sum']
            - morphology: same as neurite entry, but it will not be run on each neurite_type,
              but only once on the whole morphology.

    Returns:
        The extracted statistics

    Note:
        An example config can be found at:

    {config_path}
    """
    stats = defaultdict(dict)
    neurite_features = product(['neurite'], config.get('neurite', {}).items())
    if 'neuron' in config:    # pragma: no cover
        warn_deprecated('Usage of "neuron" is deprecated in configs of `morph_stats` package. '
                        'Use "morphology" instead.')
        config['morphology'] = config['neuron']
        del config['neuron']
    morph_features = product(['morphology'], config.get('morphology', {}).items())
    population_features = product(['population'], config.get('population', {}).items())
    neurite_types = [_NEURITE_MAP[t] for t in config.get('neurite_type', _NEURITE_MAP.keys())]

    for namespace, (feature_name, opts) in chain(neurite_features, morph_features,
                                                 population_features):
        if isinstance(opts, dict):
            kwargs = opts.get('kwargs', {})
            modes = opts.get('modes', [])
        else:
            kwargs = {}
            modes = opts
        if namespace == 'neurite':
            if 'neurite_type' not in kwargs and neurite_types:
                for t in neurite_types:
                    kwargs['neurite_type'] = t
                    stats[t.name].update(_get_feature_stats(feature_name, morphs, modes, kwargs))
            else:
                t = _NEURITE_MAP[kwargs.get('neurite_type', 'ALL')]
                kwargs['neurite_type'] = t
                stats[t.name].update(_get_feature_stats(feature_name, morphs, modes, kwargs))
        else:
            stats[namespace].update(_get_feature_stats(feature_name, morphs, modes, kwargs))

    return dict(stats)


extract_stats.__doc__ = extract_stats.__doc__.format(config_path=EXAMPLE_CONFIG)


def get_header(results):
    """Extracts the headers, using the first value in the dict as the template."""
    ret = ['name', ]
    values = next(iter(results.values()))
    for k, v in values.items():
        for metric in v.keys():
            ret.append('%s:%s' % (k, metric))
    return ret


def generate_flattened_dict(headers, results):
    """Extract from results the fields in the headers list."""
    for name, values in results.items():
        row = []
        for header in headers:
            if header == 'name':
                row.append(name)
            else:
                neurite_type, metric = header.split(':')
                row.append(values[neurite_type][metric])
        yield row


_NEURITE_MAP = {
    'AXON': nm.AXON,
    'BASAL_DENDRITE': nm.BASAL_DENDRITE,
    'APICAL_DENDRITE': nm.APICAL_DENDRITE,
    'ALL': nm.ANY_NEURITE
}


def full_config():
    """Returns a config with all features, all modes, all neurite types."""
    modes = ['min', 'max', 'median', 'mean', 'std']
    return {
        'neurite': {feature: modes for feature in _NEURITE_FEATURES},
        'morphology': {feature: modes for feature in _MORPHOLOGY_FEATURES},
        'population': {feature: modes for feature in _POPULATION_FEATURES},
        'neurite_type': list(_NEURITE_MAP.keys()),
    }


def sanitize_config(config):
    """Check that the config has the correct keys, add missing keys if necessary."""
    if 'neurite' in config:
        if 'neurite_type' not in config:
            raise ConfigError('"neurite_type" missing from config, but "neurite" set')
    else:
        config['neurite'] = {}

    if 'morphology' not in config:
        config['morphology'] = {}

    return config


def main(datapath, config, output_file, is_full_config, as_population, ignored_exceptions):
    """Main function that get statistics for morphologies.

    Args:
        datapath (str|Path): path to a morphology file or folder
        config (str|Path): path to a statistics config file
        output_file (str|Path): path to output the resulted statistics file
        is_full_config (bool): should be statistics made over all possible features, modes, neurites
        as_population (bool): treat ``datapath`` as directory of morphologies population
        ignored_exceptions (list|tuple|None): exceptions to ignore when loading a morphology
    """
    if is_full_config:
        config = full_config()
    else:
        try:
            config = get_config(config, EXAMPLE_CONFIG)
            config = sanitize_config(config)
        except ConfigError as e:
            L.error(e)
            raise

    if ignored_exceptions is None:
        ignored_exceptions = ()
    ignored_exceptions = tuple(IGNORABLE_EXCEPTIONS[k] for k in ignored_exceptions)
    morphs = nm.load_morphologies(get_files_by_path(datapath),
                                  ignored_exceptions=ignored_exceptions)

    results = {}
    if as_population:
        results[datapath] = extract_stats(morphs, config)
    else:
        for m in tqdm(morphs):
            results[m.name] = extract_stats(m, config)

    if not output_file:
        print(json.dumps(results, indent=2, separators=(',', ':'), cls=NeuromJSON))
    elif output_file.endswith('.json'):
        with open(output_file, 'w') as f:
            json.dump(results, f, cls=NeuromJSON)
    else:
        with open(output_file, 'w') as f:
            csvwriter = csv.writer(f)
            header = get_header(results)
            csvwriter.writerow(header)
            for line in generate_flattened_dict(header, dict(results)):
                csvwriter.writerow(line)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""NeuroM morphology checking functions.

Contains functions for checking validity of morphology neurites and somata.
"""
from itertools import chain, islice

import numpy as np
from neurom import NeuriteType
from neurom.check import CheckResult
from neurom.check.morphtree import get_flat_neurites
from neurom.core.morphology import Section, iter_neurites, iter_sections, iter_segments
from neurom.core.dataformat import COLS
from neurom.exceptions import NeuroMError
from neurom.morphmath import section_length, segment_length


def _read_neurite_type(neurite):
    """Simply read the stored neurite type."""
    return neurite.type


def has_axon(morph, treefun=_read_neurite_type):
    """Check if a morphology has an axon.

    Arguments:
        morph(Morphology): The morphology object to test
        treefun: Optional function to calculate the tree type of morphology's neurites

    Returns:
        CheckResult with result
    """
    return CheckResult(NeuriteType.axon in (treefun(n) for n in morph.neurites))


def has_apical_dendrite(morph, min_number=1, treefun=_read_neurite_type):
    """Check if a morphology has apical dendrites.

    Arguments:
        morph(Morphology): the morphology to test
        min_number: minimum number of apical dendrites required
        treefun: optional function to calculate the tree type of morphology's neurites

    Returns:
        CheckResult with result
    """
    types = [treefun(n) for n in morph.neurites]
    return CheckResult(types.count(NeuriteType.apical_dendrite) >= min_number)


def has_basal_dendrite(morph, min_number=1, treefun=_read_neurite_type):
    """Check if a morphology has basal dendrites.

    Arguments:
        morph(Morphology): The morphology object to test
        min_number: minimum number of basal dendrites required
        treefun: Optional function to calculate the tree type of morphology's neurites

    Returns:
        CheckResult with result
    """
    types = [treefun(n) for n in morph.neurites]
    return CheckResult(types.count(NeuriteType.basal_dendrite) >= min_number)


def has_no_flat_neurites(morph, tol=0.1, method='ratio'):
    """Check that a morphology has no flat neurites.

    Arguments:
        morph(Morphology): The morphology object to test
        tol(float): tolerance
        method(string): way of determining flatness, 'tolerance', 'ratio' \
        as described in :meth:`neurom.check.morphtree.get_flat_neurites`

    Returns:
        CheckResult with result
    """
    return CheckResult(len(get_flat_neurites(morph, tol, method)) == 0)


def has_all_nonzero_segment_lengths(morph, threshold=0.0):
    """Check presence of morphology segments with length not above threshold.

    Arguments:
        morph(Morphology): the morphology to test
        threshold(float): value above which a segment length is considered to be non-zero

    Returns:
        CheckResult with result including list of (section_id, segment_id)
        of zero length segments
    """
    bad_ids = []
    for sec in iter_sections(morph):
        p = sec.points
        for i, s in enumerate(zip(p[:-1], p[1:])):
            if segment_length(s) <= threshold:
                bad_ids.append((sec.id, i))

    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_all_nonzero_section_lengths(morph, threshold=0.0):
    """Check presence of morphology sections with length not above threshold.

    Arguments:
        morph(Morphology): the morphology to test
        threshold(float): value above which a section length is considered to be non-zero

    Returns:
        CheckResult with result including list of ids of bad sections
    """
    bad_ids = [s.id for s in iter_sections(morph.neurites)
               if section_length(s.points) <= threshold]

    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_all_nonzero_neurite_radii(morph, threshold=0.0):
    """Check presence of neurite points with radius not above threshold.

    Arguments:
        morph(Morphology): the morphology to test
        threshold(float): value above which a radius is considered to be non-zero

    Returns:
        CheckResult with result including list of (section ID, point ID) pairs
        of zero-radius points
    """
    bad_ids = []
    seen_ids = set()
    for s in iter_sections(morph):
        for i, p in enumerate(s.points):
            info = (s.id, i)
            if p[COLS.R] <= threshold and info not in seen_ids:
                seen_ids.add(info)
                bad_ids.append(info)

    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_nonzero_soma_radius(morph, threshold=0.0):
    """Check if soma radius not above threshold.

    Arguments:
        morph(Morphology): the morphology to test
        threshold: value above which the soma radius is considered to be non-zero

    Returns:
        CheckResult with result
    """
    return CheckResult(morph.soma.radius > threshold)


def has_no_jumps(morph, max_distance=30.0, axis='z'):
    """Check if there are jumps (large movements in the `axis`).

    Arguments:
        morph(Morphology): the morphology to test
        max_distance(float): value above which consecutive z-values are
        considered a jump
        axis(str): one of x/y/z, which axis to check for jumps

    Returns:
        CheckResult with result list of ids of bad sections
    """
    bad_ids = []
    axis = {'x': COLS.X, 'y': COLS.Y, 'z': COLS.Z, }[axis.lower()]
    for neurite in iter_neurites(morph):
        section_segment = ((sec, seg) for sec in iter_sections(neurite)
                           for seg in iter_segments(sec))
        for sec, (p0, p1) in islice(section_segment, 1, None):  # Skip neurite root segment
            if max_distance < abs(p0[axis] - p1[axis]):
                bad_ids.append((sec.id, [p0, p1]))
    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_no_root_node_jumps(morph, radius_multiplier=2):
    """Check that the neurites have no root node jumps.

    Their first point not should not be further than `radius_multiplier * soma radius` from the
    soma center
    """
    bad_ids = []
    for neurite in iter_neurites(morph):
        p0 = neurite.root_node.points[0, COLS.XYZ]
        distance = np.linalg.norm(p0 - morph.soma.center)
        if distance > radius_multiplier * morph.soma.radius:
            bad_ids.append((neurite.root_node.id, [p0]))
    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_no_fat_ends(morph, multiple_of_mean=2.0, final_point_count=5):
    """Check if leaf points are too large.

    Arguments:
        morph(Morphology): the morphology to test
        multiple_of_mean(float): how many times larger the final radius
        has to be compared to the mean of the final points
        final_point_count(int): how many points to include in the mean

    Returns:
        CheckResult with result list of ids of bad sections

    Note:
        A fat end is defined as a leaf segment whose last point is larger
        by a factor of `multiple_of_mean` than the mean of the points in
        `final_point_count`
    """
    bad_ids = []
    for leaf in iter_sections(morph.neurites, iterator_type=Section.ileaf):
        mean_radius = np.mean(leaf.points[1:][-final_point_count:, COLS.R])

        if mean_radius * multiple_of_mean <= leaf.points[-1, COLS.R]:
            bad_ids.append((leaf.id, leaf.points[-1:]))

    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_no_narrow_start(morph, frac=0.9):
    """Check if neurites have a narrow start.

    Arguments:
        morph(Morphology): the morphology to test
        frac(float): Ratio that the second point must be smaller than the first

    Returns:
        CheckResult with a list of all first segments of neurites with a narrow start
    """
    bad_ids = [(neurite.root_node.id, neurite.root_node.points[np.newaxis, 1])
               for neurite in morph.neurites
               if neurite.root_node.points[0][COLS.R] < frac * neurite.root_node.points[1][COLS.R]]
    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_no_dangling_branch(morph):
    """Check if the morphology has dangling neurites.

    Are considered dangling

    - dendrites whose first point is too far from the soma center
    - axons whose first point is too far from the soma center AND from
      any point belonging to a dendrite

    Arguments:
        morph(Morphology): the morphology to test

    Returns:
        CheckResult with a list of all first segments of dangling neurites
    """
    if len(morph.soma.points) == 0:
        raise NeuroMError("Can't check for dangling neurites if there is no soma")
    soma_center = morph.soma.points[:, COLS.XYZ].mean(axis=0)
    recentered_soma = morph.soma.points[:, COLS.XYZ] - soma_center
    radius = np.linalg.norm(recentered_soma, axis=1)
    soma_max_radius = radius.max()

    dendritic_points = np.array(list(chain.from_iterable(n.points
                                                         for n in iter_neurites(morph)
                                                         if n.type != NeuriteType.axon)))

    def is_dangling(neurite):
        """Is the neurite dangling?"""
        starting_point = neurite.points[0][COLS.XYZ]

        if np.linalg.norm(starting_point - soma_center) - soma_max_radius <= 12.:
            return False

        if neurite.type != NeuriteType.axon:
            return True

        distance_to_dendrites = np.linalg.norm(dendritic_points[:, COLS.XYZ] - starting_point,
                                               axis=1)
        return np.all(distance_to_dendrites >= 2 * dendritic_points[:, COLS.R] + 2)

    bad_ids = [(n.root_node.id, [n.root_node.points[0]])
               for n in iter_neurites(morph) if is_dangling(n)]
    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_no_narrow_neurite_section(morph,
                                  neurite_filter,
                                  radius_threshold=0.05,
                                  considered_section_min_length=50):
    """Check if the morphology has dendrites with narrow sections.

    Arguments:
        morph(Morphology): the morphology to test
        neurite_filter(callable): filter the neurites by this callable
        radius_threshold(float): radii below this are considered narro
        considered_section_min_length(float): sections with length below
        this are not taken into account

    Returns:
        CheckResult with result. `result.info` contains the narrow section ids and their
        first point
    """
    considered_sections = (sec for sec in iter_sections(morph, neurite_filter=neurite_filter)
                           if sec.length > considered_section_min_length)

    def narrow_section(section):
        """Select narrow sections."""
        return section.points[:, COLS.R].mean() < radius_threshold

    bad_ids = [(section.id, section.points[np.newaxis, 1])
               for section in considered_sections if narrow_section(section)]
    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_multifurcation(morph):
    """Check if a section has more than 3 children."""
    bad_ids = [(section.id, section.points[np.newaxis, -1]) for section in iter_sections(morph)
               if len(section.children) > 3]
    return CheckResult(len(bad_ids) == 0, bad_ids)


def has_no_single_children(morph):
    """Check if the morphology has sections with only one child section."""
    bad_ids = [section.id for section in iter_sections(morph) if len(section.children) == 1]
    return CheckResult(len(bad_ids) == 0, bad_ids)



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Neurite features.

Any public function from this namespace can be called via the features mechanism. If calling
directly the function in this namespace can only accept a neurite as its input. If you want to
apply it to anything other than neurite then you must use the features mechanism e.g.
``features.get``.

>>> import neurom
>>> from neurom import features
>>> m = neurom.load_morphology('path/to/morphology')
>>> features.get('max_radial_distance', m.neurites[0])
>>> features.get('max_radial_distance', m)
>>> features.get('number_of_segments', m.neurites, neurite_type=neurom.AXON)

For more details see :ref:`features`.
"""

import logging
from functools import partial
from itertools import chain

import numpy as np
import scipy
from neurom import morphmath
from neurom.core.morphology import Section
from neurom.core.dataformat import COLS
from neurom.features import NameSpace, feature, bifurcation as bf, section as sf
from neurom.geom import convex_hull
from neurom.morphmath import interval_lengths

feature = partial(feature, namespace=NameSpace.NEURITE)

L = logging.getLogger(__name__)


def _map_sections(fun, neurite, iterator_type=Section.ipreorder):
    """Map `fun` to all the sections."""
    return list(map(fun, (s for s in iterator_type(neurite.root_node))))


@feature(shape=())
def max_radial_distance(neurite):
    """Get the maximum radial distances of the termination sections."""
    term_radial_distances = section_term_radial_distances(neurite)
    return max(term_radial_distances) if term_radial_distances else 0.


@feature(shape=())
def number_of_segments(neurite):
    """Number of segments."""
    return sum(len(s.points) - 1 for s in Section.ipreorder(neurite.root_node))


@feature(shape=())
def number_of_sections(neurite, iterator_type=Section.ipreorder):
    """Number of sections. For a morphology it will be a sum of all neurites sections numbers."""
    return sum(1 for _ in iterator_type(neurite.root_node))


@feature(shape=())
def number_of_bifurcations(neurite):
    """Number of bf points."""
    return number_of_sections(neurite, iterator_type=Section.ibifurcation_point)


@feature(shape=())
def number_of_forking_points(neurite):
    """Number of forking points."""
    return number_of_sections(neurite, iterator_type=Section.iforking_point)


@feature(shape=())
def number_of_leaves(neurite):
    """Number of leaves points."""
    return number_of_sections(neurite, iterator_type=Section.ileaf)


@feature(shape=())
def total_length(neurite):
    """Neurite length. For a morphology it will be a sum of all neurite lengths."""
    return sum(s.length for s in neurite.iter_sections())


@feature(shape=())
def total_area(neurite):
    """Neurite surface area. For a morphology it will be a sum of all neurite areas.

    The area is defined as the sum of the area of the sections.
    """
    return neurite.area


@feature(shape=())
def total_volume(neurite):
    """Neurite volume. For a morphology it will be a sum of neurites volumes."""
    return sum(s.volume for s in Section.ipreorder(neurite.root_node))


def _section_length(section):
    """Get section length of `section`."""
    return morphmath.section_length(section.points)


@feature(shape=(...,))
def section_lengths(neurite):
    """Section lengths."""
    return _map_sections(_section_length, neurite)


@feature(shape=(...,))
def section_term_lengths(neurite):
    """Termination section lengths."""
    return _map_sections(_section_length, neurite, Section.ileaf)


@feature(shape=(...,))
def section_bif_lengths(neurite):
    """Bifurcation section lengths."""
    return _map_sections(_section_length, neurite, Section.ibifurcation_point)


@feature(shape=(...,))
def section_branch_orders(neurite):
    """Section branch orders."""
    return _map_sections(sf.branch_order, neurite)


@feature(shape=(...,))
def section_bif_branch_orders(neurite):
    """Bifurcation section branch orders."""
    return _map_sections(sf.branch_order, neurite, Section.ibifurcation_point)


@feature(shape=(...,))
def section_term_branch_orders(neurite):
    """Termination section branch orders."""
    return _map_sections(sf.branch_order, neurite, Section.ileaf)


@feature(shape=(...,))
def section_path_distances(neurite):
    """Path lengths."""

    def pl2(node):
        """Calculate the path length using cached section lengths."""
        return sum(n.length for n in node.iupstream())

    return _map_sections(pl2, neurite)


################################################################################
# Features returning one value per SEGMENT                                     #
################################################################################


def _map_segments(func, neurite):
    """Map `func` to all the segments.

    `func` accepts a section and returns list of values corresponding to each segment.
    """
    tmp = [mapped_seg for s in Section.ipreorder(neurite.root_node) for mapped_seg in func(s)]
    return tmp


@feature(shape=(...,))
def segment_lengths(neurite):
    """Lengths of the segments."""
    return _map_segments(sf.segment_lengths, neurite)


@feature(shape=(...,))
def segment_areas(neurite):
    """Areas of the segments."""
    return [morphmath.segment_area(seg)
            for s in Section.ipreorder(neurite.root_node)
            for seg in zip(s.points[:-1], s.points[1:])]


@feature(shape=(...,))
def segment_volumes(neurite):
    """Volumes of the segments."""

    def _func(sec):
        """List of segment volumes of a section."""
        return [morphmath.segment_volume(seg) for seg in zip(sec.points[:-1], sec.points[1:])]

    return _map_segments(_func, neurite)


@feature(shape=(...,))
def segment_radii(neurite):
    """Arithmetic mean of the radii of the points in segments."""

    def _seg_radii(sec):
        """Vectorized mean radii."""
        pts = sec.points[:, COLS.R]
        return np.divide(np.add(pts[:-1], pts[1:]), 2.0)

    return _map_segments(_seg_radii, neurite)


@feature(shape=(...,))
def segment_taper_rates(neurite):
    """Diameters taper rates of the segments.

    The taper rate is defined as the absolute radii differences divided by length of the section
    """

    def _seg_taper_rates(sec):
        """Vectorized taper rates."""
        pts = sec.points[:, COLS.XYZR]
        diff = np.diff(pts, axis=0)
        distance = np.linalg.norm(diff[:, COLS.XYZ], axis=1)
        return np.divide(2 * np.abs(diff[:, COLS.R]), distance)

    return _map_segments(_seg_taper_rates, neurite)


@feature(shape=(...,))
def section_taper_rates(neurite):
    """Diameter taper rates of the sections from root to tip.

    Taper rate is defined here as the linear fit along a section.
    It is expected to be negative for morphologies.
    """

    def _sec_taper_rate(sec):
        """Taper rate from fit along a section."""
        path_distances = np.cumsum(interval_lengths(sec.points, prepend_zero=True))
        return np.polynomial.polynomial.polyfit(path_distances, 2 * sec.points[:, COLS.R], 1)[1]

    return _map_sections(_sec_taper_rate, neurite)


@feature(shape=(...,))
def segment_meander_angles(neurite):
    """Inter-segment opening angles in a section."""
    return list(chain.from_iterable(_map_sections(sf.section_meander_angles, neurite)))


@feature(shape=(..., 3))
def segment_midpoints(neurite):
    """Return a list of segment mid-points."""

    def _seg_midpoint(sec):
        """Return the mid-points of segments in a section."""
        pts = sec.points[:, COLS.XYZ]
        return np.divide(np.add(pts[:-1], pts[1:]), 2.0)

    return _map_segments(_seg_midpoint, neurite)


@feature(shape=(...,))
def segment_path_lengths(neurite):
    """Returns pathlengths between all non-root points and their root point."""
    pathlength = {}

    def _get_pathlength(section):
        if section.id not in pathlength:
            if section.parent:
                pathlength[section.id] = section.parent.length + _get_pathlength(section.parent)
            else:
                pathlength[section.id] = 0
        return pathlength[section.id]

    result = [_get_pathlength(section) + np.cumsum(sf.segment_lengths(section))
              for section in Section.ipreorder(neurite.root_node)]
    return np.hstack(result).tolist() if result else []


@feature(shape=(...,))
def segment_radial_distances(neurite, origin=None):
    """Returns the list of distances between all segment mid points and origin."""

    def _radial_distances(sec, pos):
        """List of distances between the mid point of each segment and pos."""
        mid_pts = 0.5 * (sec.points[:-1, COLS.XYZ] + sec.points[1:, COLS.XYZ])
        return np.linalg.norm(mid_pts - pos[COLS.XYZ], axis=1)

    pos = neurite.root_node.points[0] if origin is None else origin
    # return [s for ss in n.iter_sections() for s in _radial_distances(ss, pos)]
    return [d for s in Section.ipreorder(neurite.root_node) for d in _radial_distances(s, pos)]


@feature(shape=(...,))
def local_bifurcation_angles(neurite):
    """Get a list of local bf angles."""
    return _map_sections(bf.local_bifurcation_angle,
                         neurite,
                         iterator_type=Section.ibifurcation_point)


@feature(shape=(...,))
def remote_bifurcation_angles(neurite):
    """Get a list of remote bf angles."""
    return _map_sections(bf.remote_bifurcation_angle,
                         neurite,
                         iterator_type=Section.ibifurcation_point)


@feature(shape=(...,))
def partition_asymmetry(neurite, variant='branch-order', method='petilla'):
    """Partition asymmetry at bf points.

    Variant: length is a different definition, as the absolute difference in
    downstream path lenghts, relative to the total neurite path length
    Method: 'petilla' or 'uylings'. The former is default. The latter uses ``-2`` shift. See
    :func:`neurom.features.bifurcationfunc.partition_asymmetry`
    """
    if variant not in {'branch-order', 'length'}:
        raise ValueError('Please provide a valid variant for partition asymmetry,'
                         f'found {variant}')
    if method not in {'petilla', 'uylings'}:
        raise ValueError('Please provide a valid method for partition asymmetry,'
                         'either "petilla" or "uylings"')

    if variant == 'branch-order':
        return _map_sections(
            partial(bf.partition_asymmetry, uylings=method == 'uylings'),
            neurite,
            Section.ibifurcation_point)

    asymmetries = []
    neurite_length = total_length(neurite)
    for section in Section.ibifurcation_point(neurite.root_node):
        pathlength_diff = abs(sf.downstream_pathlength(section.children[0]) -
                              sf.downstream_pathlength(section.children[1]))
        asymmetries.append(pathlength_diff / neurite_length)
    return asymmetries


@feature(shape=(...,))
def partition_asymmetry_length(neurite, method='petilla'):
    """'partition_asymmetry' feature with `variant='length'`.

    Because it is often used, it has a dedicated feature.
    """
    return partition_asymmetry(neurite, 'length', method)


@feature(shape=(...,))
def bifurcation_partitions(neurite):
    """Partition at bf points."""
    return _map_sections(bf.bifurcation_partition,
                         neurite,
                         Section.ibifurcation_point)


@feature(shape=(...,))
def sibling_ratios(neurite, method='first'):
    """Sibling ratios at bf points.

    The sibling ratio is the ratio between the diameters of the
    smallest and the largest child. It is a real number between
    0 and 1. Method argument allows one to consider mean diameters
    along the child section instead of diameter of the first point.
    """
    return _map_sections(partial(bf.sibling_ratio, method=method),
                         neurite,
                         Section.ibifurcation_point)


@feature(shape=(..., 2))
def partition_pairs(neurite):
    """Partition pairs at bf points.

    Partition pair is defined as the number of bifurcations at the two
    daughters of the bifurcating section
    """
    return _map_sections(bf.partition_pair,
                         neurite,
                         Section.ibifurcation_point)


@feature(shape=(...,))
def diameter_power_relations(neurite, method='first'):
    """Calculate the diameter power relation at a bf point.

    Diameter power relation is defined in https://www.ncbi.nlm.nih.gov/pubmed/18568015

    This quantity gives an indication of how far the branching is from
    the Rall ratio (when =1).
    """
    return _map_sections(partial(bf.diameter_power_relation, method=method),
                         neurite,
                         Section.ibifurcation_point)


@feature(shape=(...,))
def section_radial_distances(neurite, origin=None, iterator_type=Section.ipreorder):
    """Section radial distances.

    The iterator_type can be used to select only terminal sections (ileaf)
    or only bifurcations (ibifurcation_point).
    """
    pos = neurite.root_node.points[0] if origin is None else origin
    return _map_sections(partial(sf.section_radial_distance, origin=pos),
                         neurite,
                         iterator_type)


@feature(shape=(...,))
def section_term_radial_distances(neurite, origin=None):
    """Get the radial distances of the termination sections."""
    return section_radial_distances(neurite, origin, Section.ileaf)


@feature(shape=(...,))
def section_bif_radial_distances(neurite, origin=None):
    """Get the radial distances of the bf sections."""
    return section_radial_distances(neurite, origin, Section.ibifurcation_point)


@feature(shape=(...,))
def terminal_path_lengths(neurite):
    """Get the path lengths to each terminal point."""
    return _map_sections(sf.section_path_length, neurite, Section.ileaf)


@feature(shape=())
def volume_density(neurite):
    """Get the volume density.

    The volume density is defined as the ratio of the neurite volume and
    the volume of the neurite's enclosing convex hull

    TODO: convex hull fails on some morphologies, it may be good to instead use
        bounding_box to compute the neurite enclosing volume

    .. note:: Returns `np.nan` if the convex hull computation fails.
    """
    try:
        volume = convex_hull(neurite).volume
    except scipy.spatial.qhull.QhullError:
        L.exception('Failure to compute neurite volume using the convex hull. '
                    'Feature `volume_density` will return `np.nan`.\n')
        return np.nan

    return neurite.volume / volume


@feature(shape=(...,))
def section_volumes(neurite):
    """Section volumes."""
    return _map_sections(sf.section_volume, neurite)


@feature(shape=(...,))
def section_areas(neurite):
    """Section areas."""
    return _map_sections(sf.section_area, neurite)


@feature(shape=(...,))
def section_tortuosity(neurite):
    """Section tortuosities."""
    return _map_sections(sf.section_tortuosity, neurite)


@feature(shape=(...,))
def section_end_distances(neurite):
    """Section end to end distances."""
    return _map_sections(sf.section_end_distance, neurite)


@feature(shape=(...,))
def principal_direction_extents(neurite, direction=0):
    """Principal direction extent of neurites in morphologies."""
    points = neurite.points[:, :3]
    return [morphmath.principal_direction_extent(points)[direction]]


@feature(shape=(...,))
def section_strahler_orders(neurite):
    """Inter-segment opening angles in a section."""
    return _map_sections(sf.strahler_order, neurite)



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""NeuroM, lightweight and fast.

Examples:
    Obtain some morphometrics
    >>> import neurom
    >>> from neurom import features
    >>> m = neurom.load_morphology('path/to/morphology')
    >>> ap_seg_len = features.get('segment_lengths', m, neurite_type=neurom.APICAL_DENDRITE)
    >>> ax_sec_len = features.get('section_lengths', m, neurite_type=neurom.AXON)
"""
import operator
from enum import Enum
from functools import reduce

from neurom.core import Population, Morphology, Neurite
from neurom.core.morphology import iter_neurites
from neurom.core.types import NeuriteType, tree_type_checker as is_type
from neurom.exceptions import NeuroMError

_NEURITE_FEATURES = {}
_MORPHOLOGY_FEATURES = {}
_POPULATION_FEATURES = {}


class NameSpace(Enum):
    """The level of morphology abstraction that feature applies to."""
    NEURITE = 'neurite'
    NEURON = 'morphology'
    POPULATION = 'population'


def _flatten_feature(feature_shape, feature_value):
    """Flattens feature values. Applies for population features for backward compatibility."""
    if feature_shape == ():
        return feature_value
    return reduce(operator.concat, feature_value, [])


def _get_neurites_feature_value(feature_, obj, neurite_filter, kwargs):
    """Collects neurite feature values appropriately to feature's shape."""
    kwargs.pop('neurite_type', None)  # there is no 'neurite_type' arg in _NEURITE_FEATURES
    return reduce(operator.add,
                  (feature_(n, **kwargs) for n in iter_neurites(obj, filt=neurite_filter)),
                  0 if feature_.shape == () else [])


def _get_feature_value_and_func(feature_name, obj, **kwargs):
    """Obtain a feature from a set of morphology objects.

    Arguments:
        feature_name(string): feature to extract
        obj (Neurite|Morphology|Population): neurite, morphology or population
        kwargs: parameters to forward to underlying worker functions

    Returns:
        A tuple (feature, func) of the feature value and its function
    """
    # pylint: disable=too-many-branches
    is_obj_list = isinstance(obj, (list, tuple))
    if not isinstance(obj, (Neurite, Morphology, Population)) and not is_obj_list:
        raise NeuroMError('Only Neurite, Morphology, Population or list, tuple of Neurite,'
                          ' Morphology can be used for feature calculation')

    neurite_filter = is_type(kwargs.get('neurite_type', NeuriteType.all))
    res, feature_ = None, None

    if isinstance(obj, Neurite) or (is_obj_list and isinstance(obj[0], Neurite)):
        # input is a neurite or a list of neurites
        if feature_name in _NEURITE_FEATURES:
            assert 'neurite_type' not in kwargs, 'Cant apply "neurite_type" arg to a neurite with' \
                                                 ' a neurite feature'
            feature_ = _NEURITE_FEATURES[feature_name]
            if isinstance(obj, Neurite):
                res = feature_(obj, **kwargs)
            else:
                res = [feature_(s, **kwargs) for s in obj]
    elif isinstance(obj, Morphology):
        # input is a morphology
        if feature_name in _MORPHOLOGY_FEATURES:
            feature_ = _MORPHOLOGY_FEATURES[feature_name]
            res = feature_(obj, **kwargs)
        elif feature_name in _NEURITE_FEATURES:
            feature_ = _NEURITE_FEATURES[feature_name]
            res = _get_neurites_feature_value(feature_, obj, neurite_filter, kwargs)
    elif isinstance(obj, Population) or (is_obj_list and isinstance(obj[0], Morphology)):
        # input is a morphology population or a list of morphs
        if feature_name in _POPULATION_FEATURES:
            feature_ = _POPULATION_FEATURES[feature_name]
            res = feature_(obj, **kwargs)
        elif feature_name in _MORPHOLOGY_FEATURES:
            feature_ = _MORPHOLOGY_FEATURES[feature_name]
            res = _flatten_feature(feature_.shape, [feature_(n, **kwargs) for n in obj])
        elif feature_name in _NEURITE_FEATURES:
            feature_ = _NEURITE_FEATURES[feature_name]
            res = _flatten_feature(
                feature_.shape,
                [_get_neurites_feature_value(feature_, n, neurite_filter, kwargs) for n in obj])

    if res is None or feature_ is None:
        raise NeuroMError(f'Cant apply "{feature_name}" feature. Please check that it exists, '
                          'and can be applied to your input. See the features documentation page.')

    return res, feature_


def get(feature_name, obj, **kwargs):
    """Obtain a feature from a set of morphology objects.

    Features can be either Neurite, Morphology or Population features. For Neurite features see
    :mod:`neurom.features.neurite`. For Morphology features see :mod:`neurom.features.morphology`.
    For Population features see :mod:`neurom.features.population`.

    Arguments:
        feature_name(string): feature to extract
        obj: a morphology, a morphology population or a neurite tree
        kwargs: parameters to forward to underlying worker functions

    Returns:
        features as a 1D, 2D or 3D numpy array.
    """
    return _get_feature_value_and_func(feature_name, obj, **kwargs)[0]


def _register_feature(namespace: NameSpace, name, func, shape):
    """Register a feature to be applied.

    Upon registration, an attribute 'shape' containing the expected
    shape of the function return is added to 'func'.

    Arguments:
        namespace(string): a namespace, see :class:`NameSpace`
        name(string): name of the feature, used to access the feature via `neurom.features.get()`.
        func(callable): single parameter function of a neurite.
        shape(tuple): the expected shape of the feature values
    """
    setattr(func, 'shape', shape)
    _map = {NameSpace.NEURITE: _NEURITE_FEATURES,
            NameSpace.NEURON: _MORPHOLOGY_FEATURES,
            NameSpace.POPULATION: _POPULATION_FEATURES}
    if name in _map[namespace]:
        raise NeuroMError(f'A feature is already registered under "{name}"')
    _map[namespace][name] = func


def feature(shape, namespace: NameSpace, name=None):
    """Feature decorator to automatically register the feature in the appropriate namespace.

    Arguments:
        shape(tuple): the expected shape of the feature values
        namespace(string): a namespace, see :class:`NameSpace`
        name(string): name of the feature, used to access the feature via `neurom.features.get()`.
    """

    def inner(func):
        _register_feature(namespace, name or func.__name__, func, shape)
        return func

    return inner


# These imports are necessary in order to register the features
from neurom.features import neurite, morphology, \
    population  # noqa, pylint: disable=wrong-import-position



"""For backward compatibility only."""
# pylint: skip-file

from neurom.core.morphology import *  # pragma: no cover
from neurom.utils import deprecated_module  # pragma: no cover

deprecated_module('Module `neurom.core.neuron` is deprecated. Use `neurom.core.morphology`'
                  ' instead.')  # pragma: no cover



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Morphology Population Classes and Functions."""
import logging

from morphio import MorphioError
import neurom
from neurom.exceptions import NeuroMError


L = logging.getLogger(__name__)


class Population:
    """Morphology Population Class.

    Offers an iterator over morphs within population, neurites of morphs, somas of morphs.
    It does not store the loaded morphology in memory unless the morphology has been already passed
    as loaded (instance of ``Morphology``).
    """
    def __init__(self, files, name='Population', ignored_exceptions=(), cache=False):
        """Construct a morphology population.

        Arguments:
            files (collections.abc.Sequence[str|Path|Morphology]): collection of morphology files or
                paths to them or instances of ``Morphology``.
            name (str): Optional name for this Population
            ignored_exceptions (tuple): NeuroM and MorphIO exceptions that you want to ignore when
                loading morphs.
            cache (bool): whether to cache the loaded morphs in memory. If false then a morphology
                will be loaded everytime it is accessed within the population. Which is good when
                population is big. If true then all morphs will be loaded upon the construction
                and kept in memory.
        """
        self._ignored_exceptions = ignored_exceptions
        self.name = name
        if cache:
            self._files = [self._load_file(f) for f in files if f is not None]
        else:
            self._files = files

    @property
    def morphologies(self):
        """Iterator to populations's morphologies."""
        return (n for n in self)

    @property
    def somata(self):
        """Iterator to populations's somata. Somata is the plural form of soma."""
        return (n.soma for n in self)

    @property
    def neurites(self):
        """Iterator to populations's neurites."""
        return (neurite for n in self for neurite in n.neurites)

    def _load_file(self, f):
        if isinstance(f, neurom.core.morphology.Morphology):
            return f
        try:
            return neurom.load_morphology(f)
        except (NeuroMError, MorphioError) as e:
            if isinstance(e, self._ignored_exceptions):
                L.info('Ignoring exception "%s" for file %s', e, f.name)
            else:
                raise NeuroMError('`load_morphologies` failed') from e
        return None

    def __iter__(self):
        """Iterator to populations's morphs."""
        for f in self._files:
            m = self._load_file(f)
            if m is None:
                continue
            yield m

    def __len__(self):
        """Length of morphology collection."""
        return len(self._files)

    def __getitem__(self, idx):
        """Get morphology at index idx."""
        if idx > len(self):
            raise ValueError(
                f'no {idx} index in "{self.name}" population, max possible index is {len(self)}')
        return self._load_file(self._files[idx])

    def __str__(self):
        """Return a string representation."""
        return f'Population <name: {self.name}, n_morphologies: {len(self)}>'



"""For backward compatibility only."""
# pylint: skip-file

from neurom.features.bifurcation import *  # pragma: no cover
from neurom.utils import deprecated_module  # pragma: no cover

deprecated_module('Module `neurom.features.bifurcationfunc` is deprecated. Use'
                  '`neurom.features.bifurcation` instead.')  # pragma: no cover



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Population features.

Any public function from this namespace can be called via features mechanism. Functions in this
namespace can only accept a morphology population as its input no matter how called.

>>> import neurom
>>> from neurom import features
>>> pop = neurom.load_morphologies('path/to/morphs')
>>> features.get('sholl_frequency', pop)

For more details see :ref:`features`.
"""


from functools import partial
import numpy as np

from neurom.core.dataformat import COLS
from neurom.core.types import NeuriteType
from neurom.core.types import tree_type_checker as is_type
from neurom.features import feature, NameSpace
from neurom.features.morphology import sholl_crossings

feature = partial(feature, namespace=NameSpace.POPULATION)


@feature(shape=(...,))
def sholl_frequency(morphs, neurite_type=NeuriteType.all, step_size=10, bins=None):
    """Perform Sholl frequency calculations on a population of morphs.

    Args:
        morphs(list|Population): list of morphologies or morphology population
        neurite_type(NeuriteType): which neurites to operate on
        step_size(float): step size between Sholl radii
        bins(iterable of floats): custom binning to use for the Sholl radii. If None, it uses
        intervals of step_size between min and max radii of ``morphs``.

    Note:
        Given a population, the concentric circles range from the smallest soma radius to the
        largest radial neurite distance in steps of `step_size`. Each segment of the morphology is
        tested, so a neurite that bends back on itself, and crosses the same Sholl radius will
        get counted as having crossed multiple times.
    """
    neurite_filter = is_type(neurite_type)

    if bins is None:
        min_soma_edge = min(n.soma.radius for n in morphs)
        max_radii = max(np.max(np.linalg.norm(n.points[:, COLS.XYZ], axis=1))
                        for m in morphs
                        for n in m.neurites if neurite_filter(n))
        bins = np.arange(min_soma_edge, min_soma_edge + max_radii, step_size)

    return sum(np.array(sholl_crossings(m, m.soma.center, bins, neurite_type))
               for m in morphs)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Check on morphologies."""

import json
from pathlib import Path
import pkg_resources
from neurom.apps import get_config
from neurom.check.runner import CheckRunner

EXAMPLE_CONFIG = Path(pkg_resources.resource_filename('neurom.apps', 'config'), 'morph_check.yaml')


def main(datapath, config, output):
    """Main function that checks morphologies.

    Args:
        datapath (str|Path): path to a morphology file or folder
        config (str|Path): path to a statistics config file
        output (str|Path): path to output the resulted checks file
    """
    config = get_config(config, EXAMPLE_CONFIG)
    checker = CheckRunner(config)
    summary = checker.run(datapath)
    with open(output, 'w') as json_output:
        json.dump(summary, json_output, indent=4)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 501ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Morphology draw functions using matplotlib."""

from functools import wraps
import numpy as np
from matplotlib.collections import LineCollection, PatchCollection
from matplotlib.lines import Line2D
from matplotlib.patches import Circle, FancyArrowPatch, Polygon, Rectangle
from mpl_toolkits.mplot3d.art3d import Line3DCollection
from neurom import NeuriteType, geom
from neurom.core.morphology import iter_neurites, iter_sections, iter_segments
from neurom.core.soma import SomaCylinders
from neurom.core.dataformat import COLS
from neurom.core.types import tree_type_checker
from neurom.morphmath import segment_radius
from neurom.view.dendrogram import Dendrogram, get_size, layout_dendrogram, move_positions

from neurom.view import matplotlib_utils

_LINEWIDTH = 1.2
_ALPHA = 0.8
_DIAMETER_SCALE = 1.0
TREE_COLOR = {NeuriteType.basal_dendrite: 'red',
              NeuriteType.apical_dendrite: 'purple',
              NeuriteType.axon: 'blue',
              NeuriteType.soma: 'black',
              NeuriteType.undefined: 'green',
              NeuriteType.custom5: 'orange',
              NeuriteType.custom6: 'orange',
              NeuriteType.custom7: 'orange',
              NeuriteType.custom8: 'orange',
              NeuriteType.custom9: 'orange',
              NeuriteType.custom10: 'orange'}


def _implicit_ax(plot_func, params=None):
    """Sets ``ax`` arg for plot functions if ``ax`` is not set originally."""

    @wraps(plot_func)
    def wrapper(*args, **kwargs):
        fig = None
        ax = kwargs.get('ax', None)
        if ax is None and len(args) == 1:
            fig, ax = matplotlib_utils.get_figure(params=params)
            kwargs['ax'] = ax
        res = plot_func(*args, **kwargs)
        if fig:
            matplotlib_utils.plot_style(fig=fig, ax=ax)
        return res

    return wrapper


def _implicit_ax3d(plot_func):
    return _implicit_ax(plot_func, {'projection': '3d'})


def _plane2col(plane):
    """Take a string like 'xy', and return the indices from COLS.*."""
    planes = ('xy', 'yx', 'xz', 'zx', 'yz', 'zy')
    assert plane in planes, 'No such plane found! Please select one of: ' + str(planes)
    return (getattr(COLS, plane[0].capitalize()),
            getattr(COLS, plane[1].capitalize()), )


def _get_linewidth(tree, linewidth, diameter_scale):
    """Calculate the desired linewidth based on tree contents.

    If diameter_scale exists, it is used to scale the diameter of each of the segments
    in the tree
    If diameter_scale is None, the linewidth is used.
    """
    if diameter_scale is not None and tree:
        linewidth = [2 * segment_radius(s) * diameter_scale
                     for s in iter_segments(tree)]
    return linewidth


def _get_color(treecolor, tree_type):
    """If treecolor set, it's returned, otherwise tree_type is used to return set colors."""
    if treecolor is not None:
        return treecolor
    return TREE_COLOR.get(tree_type, 'green')


@_implicit_ax
def plot_tree(tree, ax=None, plane='xy',
              diameter_scale=_DIAMETER_SCALE, linewidth=_LINEWIDTH,
              color=None, alpha=_ALPHA, realistic_diameters=False):
    """Plots a 2d figure of the tree's segments.

    Args:
        tree(neurom.core.Section or neurom.core.Neurite): plotted tree
        ax(matplotlib axes): on what to plot
        plane(str): Any pair of 'xyz'
        diameter_scale(float): Scale factor multiplied with segment diameters before plotting
        linewidth(float): all segments are plotted with this width, but only if diameter_scale=None
        color(str or None): Color of plotted values, None corresponds to default choice
        alpha(float): Transparency of plotted values
        realistic_diameters(bool): scale linewidths with axis data coordinates

    Note:
        If the tree contains one single point the plot will be empty
        since no segments can be constructed.
    """
    plane0, plane1 = _plane2col(plane)

    section_segment_list = [(section, segment)
                            for section in iter_sections(tree)
                            for segment in iter_segments(section)]
    colors = [_get_color(color, section.type) for section, _ in section_segment_list]

    if realistic_diameters:
        def _get_rectangle(x, y, linewidth):
            """Draw  a rectangle to represent a secgment."""
            x, y = np.array(x), np.array(y)
            diff = y - x
            angle = np.arctan2(diff[1], diff[0]) % (2 * np.pi)
            return Rectangle(x - linewidth / 2. * np.array([-np.sin(angle), np.cos(angle)]),
                             np.linalg.norm(diff),
                             linewidth,
                             np.rad2deg(angle))

        segs = [_get_rectangle((seg[0][plane0], seg[0][plane1]),
                               (seg[1][plane0], seg[1][plane1]),
                               2 * segment_radius(seg) * diameter_scale)
                for _, seg in section_segment_list]

        collection = PatchCollection(segs, alpha=alpha, facecolors=colors)

    else:
        segs = [((seg[0][plane0], seg[0][plane1]),
                 (seg[1][plane0], seg[1][plane1]))
                for _, seg in section_segment_list]

        linewidth = _get_linewidth(
            tree,
            diameter_scale=diameter_scale,
            linewidth=linewidth,
        )
        collection = LineCollection(segs, colors=colors, linewidth=linewidth, alpha=alpha)

    ax.add_collection(collection)


@_implicit_ax
def plot_soma(soma, ax=None, plane='xy',
              soma_outline=True,
              linewidth=_LINEWIDTH,
              color=None, alpha=_ALPHA):
    """Generates a 2d figure of the soma.

    Args:
        soma(neurom.core.Soma): plotted soma
        ax(matplotlib axes): on what to plot
        plane(str): Any pair of 'xyz'
        soma_outline(bool): should the soma be drawn as an outline
        linewidth(float): all segments are plotted with this width, but only if diameter_scale=None
        color(str or None): Color of plotted values, None corresponds to default choice
        alpha(float): Transparency of plotted values
    """
    plane0, plane1 = _plane2col(plane)
    color = _get_color(color, tree_type=NeuriteType.soma)

    if isinstance(soma, SomaCylinders):
        for start, end in zip(soma.points, soma.points[1:]):
            matplotlib_utils.project_cylinder_onto_2d(
                ax, (plane0, plane1),
                start=start[COLS.XYZ], end=end[COLS.XYZ],
                start_radius=start[COLS.R], end_radius=end[COLS.R],
                color=color, alpha=alpha)
    else:
        if soma_outline:
            ax.add_artist(Circle(soma.center[[plane0, plane1]], soma.radius,
                                 color=color, alpha=alpha))
        else:
            points = [[p[plane0], p[plane1]] for p in soma.iter()]
            if points:
                points.append(points[0])  # close the loop
                x, y = tuple(np.array(points).T)
                ax.plot(x, y, color=color, alpha=alpha, linewidth=linewidth)

    ax.set_xlabel(plane[0])
    ax.set_ylabel(plane[1])

    bounding_box = geom.bounding_box(soma)
    ax.dataLim.update_from_data_xy(np.vstack(([bounding_box[0][plane0], bounding_box[0][plane1]],
                                              [bounding_box[1][plane0], bounding_box[1][plane1]])),
                                   ignore=False)


# pylint: disable=too-many-arguments
@_implicit_ax
def plot_morph(morph, ax=None,
               neurite_type=NeuriteType.all,
               plane='xy',
               soma_outline=True,
               diameter_scale=_DIAMETER_SCALE, linewidth=_LINEWIDTH,
               color=None, alpha=_ALPHA, realistic_diameters=False):
    """Plots a 2D figure of the morphology, that contains a soma and the neurites.

    Args:
        neurite_type(NeuriteType|tuple): an optional filter on the neurite type
        ax(matplotlib axes): on what to plot
        morph(Morphology): morphology to be plotted
        soma_outline(bool): should the soma be drawn as an outline
        plane(str): Any pair of 'xyz'
        diameter_scale(float): Scale factor multiplied with segment diameters before plotting
        linewidth(float): all segments are plotted with this width, but only if diameter_scale=None
        color(str or None): Color of plotted values, None corresponds to default choice
        alpha(float): Transparency of plotted values
        realistic_diameters(bool): scale linewidths with axis data coordinates
    """
    plot_soma(morph.soma, ax, plane=plane, soma_outline=soma_outline, linewidth=linewidth,
              color=color, alpha=alpha)

    for neurite in iter_neurites(morph, filt=tree_type_checker(neurite_type)):
        plot_tree(neurite, ax, plane=plane,
                  diameter_scale=diameter_scale, linewidth=linewidth,
                  color=color, alpha=alpha, realistic_diameters=realistic_diameters)

    ax.set_title(morph.name)
    ax.set_xlabel(plane[0])
    ax.set_ylabel(plane[1])


def _update_3d_datalim(ax, obj):
    """Unlike w/ 2d Axes, the dataLim isn't set by collections, so it has to be updated manually."""
    min_bounding_box, max_bounding_box = geom.bounding_box(obj)
    xy_bounds = np.vstack((min_bounding_box[:COLS.Z],
                           max_bounding_box[:COLS.Z]))
    ax.xy_dataLim.update_from_data_xy(xy_bounds, ignore=False)

    z_bounds = np.vstack(((min_bounding_box[COLS.Z], min_bounding_box[COLS.Z]),
                          (max_bounding_box[COLS.Z], max_bounding_box[COLS.Z])))
    ax.zz_dataLim.update_from_data_xy(z_bounds, ignore=False)


@_implicit_ax3d
def plot_tree3d(tree, ax=None,
                diameter_scale=_DIAMETER_SCALE, linewidth=_LINEWIDTH,
                color=None, alpha=_ALPHA):
    """Generates a figure of the tree in 3d.

    If the tree contains one single point the plot will be empty \
    since no segments can be constructed.

    Args:
        tree(neurom.core.Section or neurom.core.Neurite): plotted tree
        ax(matplotlib axes): on what to plot
        diameter_scale(float): Scale factor multiplied with segment diameters before plotting
        linewidth(float): all segments are plotted with this width, but only if diameter_scale=None
        color(str or None): Color of plotted values, None corresponds to default choice
        alpha(float): Transparency of plotted values
    """
    section_segment_list = [(section, segment)
                            for section in iter_sections(tree)
                            for segment in iter_segments(section)]
    segs = [(seg[0][COLS.XYZ], seg[1][COLS.XYZ]) for _, seg in section_segment_list]
    colors = [_get_color(color, section.type) for section, _ in section_segment_list]

    linewidth = _get_linewidth(tree, diameter_scale=diameter_scale, linewidth=linewidth)

    collection = Line3DCollection(segs, colors=colors, linewidth=linewidth, alpha=alpha)
    ax.add_collection3d(collection)

    _update_3d_datalim(ax, tree)


@_implicit_ax3d
def plot_soma3d(soma, ax=None, color=None, alpha=_ALPHA):
    """Generates a 3d figure of the soma.

    Args:
        soma(neurom.core.Soma): plotted soma
        ax(matplotlib axes): on what to plot
        color(str or None): Color of plotted values, None corresponds to default choice
        alpha(float): Transparency of plotted values
    """
    color = _get_color(color, tree_type=NeuriteType.soma)

    if isinstance(soma, SomaCylinders):
        for start, end in zip(soma.points, soma.points[1:]):
            matplotlib_utils.plot_cylinder(ax,
                                           start=start[COLS.XYZ], end=end[COLS.XYZ],
                                           start_radius=start[COLS.R], end_radius=end[COLS.R],
                                           color=color, alpha=alpha)
    else:
        matplotlib_utils.plot_sphere(ax, center=soma.center[COLS.XYZ], radius=soma.radius,
                                     color=color, alpha=alpha)

    # unlike w/ 2d Axes, the dataLim isn't set by collections, so it has to be updated manually
    _update_3d_datalim(ax, soma)


@_implicit_ax3d
def plot_morph3d(morph, ax=None, neurite_type=NeuriteType.all,
                 diameter_scale=_DIAMETER_SCALE, linewidth=_LINEWIDTH,
                 color=None, alpha=_ALPHA):
    """Generates a figure of the morphology, that contains a soma and a list of trees.

    Args:
        morph(Morphology): morphology to be plotted
        ax(matplotlib axes): on what to plot
        neurite_type(NeuriteType): an optional filter on the neurite type
        diameter_scale(float): Scale factor multiplied with segment diameters before plotting
        linewidth(float): all segments are plotted with this width, but only if diameter_scale=None
        color(str or None): Color of plotted values, None corresponds to default choice
        alpha(float): Transparency of plotted values
    """
    plot_soma3d(morph.soma, ax, color=color, alpha=alpha)

    for neurite in iter_neurites(morph, filt=tree_type_checker(neurite_type)):
        plot_tree3d(neurite, ax,
                    diameter_scale=diameter_scale, linewidth=linewidth,
                    color=color, alpha=alpha)

    ax.set_title(morph.name)


def _get_dendrogram_legend(dendrogram):
    """Generates labels legend for dendrogram.

    Because dendrogram is rendered as patches, we need to manually label it.
    Args:
        dendrogram (Dendrogram): dendrogram

    Returns:
        List of legend handles.
    """
    def neurite_legend(neurite_type):
        return Line2D([0], [0], color=TREE_COLOR[neurite_type], lw=2, label=neurite_type.name)

    if dendrogram.neurite_type == NeuriteType.soma:
        handles = {d.neurite_type: neurite_legend(d.neurite_type)
                   for d in [dendrogram] + dendrogram.children}
        return handles.values()
    return [neurite_legend(dendrogram.neurite_type)]


def _as_dendrogram_polygon(coords, color):
    return Polygon(coords, color=color, fill=True)


def _as_dendrogram_line(start, end, color):
    return FancyArrowPatch(start, end, arrowstyle='-', color=color, lw=2, shrinkA=0, shrinkB=0)


def _get_dendrogram_shapes(dendrogram, positions, show_diameters):
    """Generates drawable patches for dendrogram.

    Args:
        dendrogram (Dendrogram): dendrogram
        positions (dict of Dendrogram: np.array): positions xy coordinates of dendrograms
        show_diameter (bool): whether to draw shapes with diameter or as plain lines

    Returns:
        List of matplotlib.patches.
    """
    color = TREE_COLOR[dendrogram.neurite_type]
    start_point = positions[dendrogram]
    end_point = start_point + [0, dendrogram.height]
    if show_diameters:
        shapes = [_as_dendrogram_polygon(dendrogram.coords + start_point, color)]
    else:
        shapes = [_as_dendrogram_line(start_point, end_point, color)]
    for child in dendrogram.children:
        shapes.append(_as_dendrogram_line(end_point, positions[child], color))
        shapes += _get_dendrogram_shapes(child, positions, show_diameters)
    return shapes


@_implicit_ax
def plot_dendrogram(obj, ax=None, show_diameters=True):
    """Plots Dendrogram of `obj`.

    Args:
        obj (neurom.Morphology, neurom.Section): morphology or section
        ax: matplotlib axes
        show_diameters (bool): whether to show node diameters or not
    """
    dendrogram = Dendrogram(obj)
    positions = layout_dendrogram(dendrogram, np.array([0, 0]))
    w, h = get_size(positions)
    positions = move_positions(positions, np.array([.5 * w, 0]))
    ax.set_xlim([-.05 * w, 1.05 * w])
    ax.set_ylim([-.05 * h, 1.05 * h])
    ax.set_title('Morphology Dendrogram')
    ax.set_xlabel('micrometers (um)')
    ax.set_ylabel('micrometers (um)')
    shapes = _get_dendrogram_shapes(dendrogram, positions, show_diameters)
    ax.add_collection(PatchCollection(shapes, match_original=True))

    ax.set_aspect('auto')
    ax.legend(handles=_get_dendrogram_legend(dendrogram))



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Morphology features.

Any public function from this namespace can be called via the features mechanism. If calling
directly the function in this namespace can only accept a morphology as its input. If you want to
apply it to a morphology population then you must use the features mechanism e.g. ``features.get``.
The features mechanism does not allow you to apply these features to neurites.

>>> import neurom
>>> from neurom import features
>>> m = neurom.load_morphology('path/to/morphology')
>>> features.get('soma_surface_area', m)
>>> population = neurom.load_morphologies('path/to/morphs')
>>> features.get('sholl_crossings', population)

For more details see :ref:`features`.
"""


from functools import partial
import math
import numpy as np

from neurom import morphmath
from neurom.core.morphology import iter_neurites, iter_segments, Morphology
from neurom.core.types import tree_type_checker as is_type
from neurom.core.dataformat import COLS
from neurom.core.types import NeuriteType
from neurom.features import feature, NameSpace, neurite as nf

feature = partial(feature, namespace=NameSpace.NEURON)


@feature(shape=())
def soma_volume(morph):
    """Get the volume of a morphology's soma."""
    return morph.soma.volume


@feature(shape=())
def soma_surface_area(morph):
    """Get the surface area of a morphology's soma.

    Note:
        The surface area is calculated by assuming the soma is spherical.
    """
    return 4 * math.pi * morph.soma.radius ** 2


@feature(shape=())
def soma_radius(morph):
    """Get the radius of a morphology's soma."""
    return morph.soma.radius


@feature(shape=())
def max_radial_distance(morph, neurite_type=NeuriteType.all):
    """Get the maximum radial distances of the termination sections."""
    term_radial_distances = [nf.max_radial_distance(n)
                             for n in iter_neurites(morph, filt=is_type(neurite_type))]
    return max(term_radial_distances) if term_radial_distances else 0.


@feature(shape=(...,))
def number_of_sections_per_neurite(morph, neurite_type=NeuriteType.all):
    """List of numbers of sections per neurite."""
    return [nf.number_of_sections(n)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def total_length_per_neurite(morph, neurite_type=NeuriteType.all):
    """Neurite lengths."""
    return [nf.total_length(n)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def total_area_per_neurite(morph, neurite_type=NeuriteType.all):
    """Neurite areas."""
    return [nf.total_area(n)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def total_volume_per_neurite(morph, neurite_type=NeuriteType.all):
    """Neurite volumes."""
    return [nf.total_volume(n)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def trunk_origin_azimuths(morph, neurite_type=NeuriteType.all):
    """Get a list of all the trunk origin azimuths of a morph.

    The azimuth is defined as Angle between x-axis and the vector
    defined by (initial tree point - soma center) on the x-z plane.

    The range of the azimuth angle [-pi, pi] radians
    """
    def _azimuth(section, soma):
        """Azimuth of a section."""
        vector = morphmath.vector(section[0], soma.center)
        return np.arctan2(vector[COLS.Z], vector[COLS.X])

    return [_azimuth(n.root_node.points, morph.soma)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def trunk_origin_elevations(morph, neurite_type=NeuriteType.all):
    """Get a list of all the trunk origin elevations of a morph.

    The elevation is defined as the angle between x-axis and the
    vector defined by (initial tree point - soma center)
    on the x-y half-plane.

    The range of the elevation angle [-pi/2, pi/2] radians
    """
    def _elevation(section, soma):
        """Elevation of a section."""
        vector = morphmath.vector(section[0], soma.center)
        norm_vector = np.linalg.norm(vector)

        if norm_vector >= np.finfo(type(norm_vector)).eps:
            return np.arcsin(vector[COLS.Y] / norm_vector)
        raise ValueError("Norm of vector between soma center and section is almost zero.")

    return [_elevation(n.root_node.points, morph.soma)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def trunk_vectors(morph, neurite_type=NeuriteType.all):
    """Calculate the vectors between all the trunks of the morphology and the soma center."""
    return [morphmath.vector(n.root_node.points[0], morph.soma.center)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def trunk_angles(morph, neurite_type=NeuriteType.all):
    """Calculate the angles between all the trunks of the morph.

    The angles are defined on the x-y plane and the trees
    are sorted from the y axis and anticlock-wise.
    """
    vectors = np.array(trunk_vectors(morph, neurite_type=neurite_type))
    # In order to avoid the failure of the process in case the neurite_type does not exist
    if len(vectors) == 0:
        return []

    def _sort_angle(p1, p2):
        """Angle between p1-p2 to sort vectors."""
        ang1 = np.arctan2(*p1[::-1])
        ang2 = np.arctan2(*p2[::-1])
        return ang1 - ang2

    # Sorting angles according to x-y plane
    order = np.argsort(np.array([_sort_angle(i / np.linalg.norm(i), [0, 1])
                                 for i in vectors[:, 0:2]]))

    ordered_vectors = vectors[order][:, [COLS.X, COLS.Y]]

    return [morphmath.angle_between_vectors(ordered_vectors[i], ordered_vectors[i - 1])
            for i, _ in enumerate(ordered_vectors)]


@feature(shape=(...,))
def trunk_origin_radii(morph, neurite_type=NeuriteType.all):
    """Radii of the trunk sections of neurites in a morph."""
    return [n.root_node.points[0][COLS.R]
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def trunk_section_lengths(morph, neurite_type=NeuriteType.all):
    """List of lengths of trunk sections of neurites in a morph."""
    return [morphmath.section_length(n.root_node.points)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=())
def number_of_neurites(morph, neurite_type=NeuriteType.all):
    """Number of neurites in a morph."""
    return sum(1 for _ in iter_neurites(morph, filt=is_type(neurite_type)))


@feature(shape=(...,))
def neurite_volume_density(morph, neurite_type=NeuriteType.all):
    """Get volume density per neurite."""
    return [nf.volume_density(n)
            for n in iter_neurites(morph, filt=is_type(neurite_type))]


@feature(shape=(...,))
def sholl_crossings(morph, center=None, radii=None, neurite_type=NeuriteType.all):
    """Calculate crossings of neurites.

    Args:
        morph(Morphology|list): morphology or a list of neurites
        center(Point): center point, if None then soma center is taken
        radii(iterable of floats): radii for which crossings will be counted,
            if None then soma radius is taken
        neurite_type(NeuriteType): Type of neurite to use. By default ``NeuriteType.all`` is used.

    Returns:
        Array of same length as radii, with a count of the number of crossings
        for the respective radius

    This function can also be used with a list of sections, as follow::

        secs = (sec for sec in nm.iter_sections(morph) if complex_filter(sec))
        sholl = nm.features.neuritefunc.sholl_crossings(secs,
                                                        center=morph.soma.center,
                                                        radii=np.arange(0, 1000, 100))
    """
    def _count_crossings(neurite, radius):
        """Used to count_crossings of segments in neurite with radius."""
        r2 = radius ** 2
        count = 0
        for start, end in iter_segments(neurite):
            start_dist2, end_dist2 = (morphmath.point_dist2(center, start),
                                      morphmath.point_dist2(center, end))

            count += int(start_dist2 <= r2 <= end_dist2 or
                         end_dist2 <= r2 <= start_dist2)

        return count

    if center is None or radii is None:
        assert isinstance(morph, Morphology) and morph.soma, \
            '`sholl_crossings` input error. If `center` or `radii` is not set then `morph` is ' \
            'expected to be an instance of Morphology and have a soma.'
        if center is None:
            center = morph.soma.center
        if radii is None:
            radii = [morph.soma.radius]
    return [sum(_count_crossings(neurite, r)
                for neurite in iter_neurites(morph, filt=is_type(neurite_type)))
            for r in radii]


@feature(shape=(...,))
def sholl_frequency(morph, neurite_type=NeuriteType.all, step_size=10, bins=None):
    """Perform Sholl frequency calculations on a morph.

    Args:
        morph(Morphology): a morphology
        neurite_type(NeuriteType): which neurites to operate on
        step_size(float): step size between Sholl radii
        bins(iterable of floats): custom binning to use for the Sholl radii. If None, it uses
        intervals of step_size between min and max radii of ``morphologies``.

    Note:
        Given a morphology, the soma center is used for the concentric circles,
        which range from the soma radii, and the maximum radial distance
        in steps of `step_size`. Each segment of the morphology is tested, so a neurite that
        bends back on itself, and crosses the same Sholl radius will get counted as
        having crossed multiple times.
    """
    neurite_filter = is_type(neurite_type)

    if bins is None:
        min_soma_edge = morph.soma.radius
        max_radii = max(np.max(np.linalg.norm(n.points[:, COLS.XYZ], axis=1))
                        for n in morph.neurites if neurite_filter(n))
        bins = np.arange(min_soma_edge, min_soma_edge + max_radii, step_size)

    return sholl_crossings(morph, morph.soma.center, bins, neurite_type)
